// copyright 2012–2021 nick black
#ifndef GROWLIGHT_SRC_NOTUI_AGGREGATE
#define GROWLIGHT_SRC_NOTUI_AGGREGATE

#ifdef __cplusplus
extern "C" {
#endif

int raise_aggregate_form(void);

#ifdef __cplusplus
}
#endif

#endif
