// copyright 2012–2021 nick black
#ifndef GROWLIGHT_DM
#define GROWLIGHT_DM

#ifdef __cplusplus
extern "C" {
#endif

struct device;

int explore_dm_sysfs(struct device *,int);

#ifdef __cplusplus
}
#endif

#endif
